"""
    Cursor constants are used for server side cursors.
    Currently only read only cursor is supported.

    Cursor constants are defined in module *mariadb.constants.CURSOR*.
"""

NONE= 0
READ_ONLY= 1
