"""
Constants for _get_info method of MariadB connection object
"""

CHARSET_ID= 0
CHARSET_NAME= 1 
CLIENT_ERRORS= 2
CLIENT_VERSION= 3
CLIENT_VERSION_ID= 4
ASYNC_TIMEOUT= 5
ASYNC_TIMEOUT_MS= 6
CHARSET_INFO= 7
ERROR= 8
ERROR_ID= 9
HOST= 10
INFO= 11
PORT= 12
PROTOCOL_VERSION_ID= 13
PVIO_TYPE= 14
SCHEMA= 15
SERVER_TYPE= 16
SERVER_VERSION= 17 
SERVER_VERSION_ID= 18 
SOCKET= 19
SQLSTATE= 20
SSL_CIPHER= 21
TLS_LIBRARY= 22
TLS_VERSION= 23
TLS_VERSION_ID= 24
TYPE= 25
UNIX_SOCKET= 26
USER= 27
MAX_ALLOWED_PACKET= 28
NET_BUFFER_LENGTH= 29
SERVER_STATUS= 30
SERVER_CAPABILITIES= 31
EXTENDED_SERVER_CAPABILITIES= 32
CLIENT_CAPABILITIES= 33
BYTES_READ= 34
BYTES_SENT= 35
